# Simulation EM Acceleration ------------------------------------------------------------------

# Code by Ortrud Wartlick and Marie Beisemann

# ---------------------------------------------------------------------------------------------

# set working directory to script location (works only when script is sourced!) ---------------
# setwd(getSrcDirectory()[1])

# load required packages ----------------------------------------------------------------------

if (!require("turboEM")) {
  install.packages("turboEM")
}

if (!require("mirt")) {
  install.packages("mirt")
}
# in order for this script to run properly, mirt version 1.31 is required
# newer versions have had internal changes which we cause issues with this script

if (!require("mvtnorm")) {
  install.packages("mvtnorm")
}

library(turboEM)
library(mirt)
library(mvtnorm)
library(doParallel)
library(doRNG)

# functions ---------------------------------------------------------------------------------

# norm_vec ----------------------------------------------------------------------------------
# Auxiliary function to calculate euclidian vector norms.
# If x is a matrix, the norm is calculated for every column.
norm_vec <- function(x){
    apply(x,2,function(x) sqrt(sum(x^2)))
}

# getTrajectories --------------------------------------------------------------------------
# Obtain trajectories, starting from different initial values contained as
# columns in input pInits. Abort when converged or when maximum number of
# iterations is achieved.
# @param pInits Each column of pInits contains initial starting values.
# @param x The data.
# @param fixptnfun The fixed point function F.
# @param maxIter The maximum number of iterations.
# @param ... Additional parameters for the fixed point function.
getTrajectories <- function(pInits, x, fixptnfun, maxIter = 1000, ...){
  trajectories = list()
  # Calculate a trajectory for each starting value, i.e. each column in pInits.
  for (init in seq_along(pInits[1,])){
    print(paste0('Trajectory: ', init))
    # Preallocate space for sequence of parameter estimates.
    p <- matrix(rep(NaN, maxIter*dim(pInits)[1]), ncol = maxIter)
    p[,1] <- pInits[,init]
    STOP <- FALSE
    i <- 1
    # Run standard EM until convergence.
    while (!STOP && i < maxIter){
      i <- i + 1
      p[,i] <- fixptnfun(p[,i-1], x, ...)
      # Check for convergence.
      STOP <- sqrt(crossprod(p[,i] - p[,i-1])) < 10^(-7)
      }
    # Add trajectory to list:
    trajectories[[init]] <- p
    }
  trajectories
}

# r ----------------------------------------------------------------------------------------------
# Properties of F
# The method r calculates the relative distance between successive elements in a sequence of thetas (column-wise).
# The steplength is computed relative to the total distance traveled between the starting theta and the fixed point.
# The method returns a vector with ncol(sequence)-1 elements.
# The input sequence must be a matrix.
r <- function(sequence){
    # Total number of iterates.
    Nt <- length(sequence[1,])
    # F(theta)-theta for Nt-1 elements of the sequence of thetas (column-wise).
    U <- matrix(sequence[,2:Nt] - sequence[,1:Nt-1], ncol = Nt-1)
    # Return the relative distances.
    absDistances <- norm_vec(U)
    absDistances / sum(absDistances, na.rm = TRUE)
}

# psi --------------------------------------------------------------------------------------------
# Properties of F
# The method psi calculates the angles between successive iterates in a sequence of thetas (column-wise).
# It returns a vector with ncol(sequence)-2 elements.
# The input sequence must be a matrix.
psi <- function(sequence){
    # Total number of iterates.
    Nt <- length(sequence[1,])
    # F(theta)-theta for Nt-1 elements of the sequence of thetas (column-wise).
    U <- matrix(sequence[,2:Nt] - sequence[,1:(Nt-1)], ncol = Nt-1)
    # F(F(theta))-F(theta) for Nt-2 elements of the sequence of thetas (column-wise).
    V <- matrix(U[,-1], ncol = Nt-2)
    # Remove last column of U to match dimensions.
    U <- matrix(U[,-(Nt-1)], ncol = Nt-2)
    # Return the angles betwen each column of U and V.
    acos( diag(t(U) %*% V) / (norm_vec(U) * norm_vec(V)))
}

# repackageParams ---------------------------------------------------------------------------------
# Repackage a vector of IRT parameters with additional information into a dataframe
# that is accepted as input by mirt.
# Possible paramIds are: 'delta', 'alpha' and 'guessing'. (Upperbounds are not considered at the moment.)
# The last entry in p is the variance. It is estimated for Rasch models but not for 2PL models.
repackageParams <- function(p, x, paramIds, estimate_var = TRUE){
    itemsAGroup <- c(rep(colnames(x),each = 4), rep("GROUP",2))
    names <- c(rep(c('a1','d','g','u'),times = ncol(x)), 'MEAN_1', 'COV_11')
    values <- rep(NaN, length(itemsAGroup))
    for (j in seq(from = 1, to = ncol(x), by = 1)){
        if ('delta' %in% paramIds){
            values[4*j-2] <- p[j]
        } else {
            values[4*j-2] <- 1
        }
        if ('alpha' %in% paramIds){
            values[4*j-3] <- p[ncol(x) + j]
        } else {
            values[4*j-3] <- 1
        }
        if ('guessing' %in% paramIds){
            values[4*j-1] <- p[2*ncol(x) +j]
        } else {
            values[4*j-1] <- 0
        }
        values[4*j] <- 1
    }
    # mean ability is assumed to be zero.
    values[length(values)-1] <- 0
    if (estimate_var){
        # estimate variance
        values[length(values)] <- p[length(p)]
    } else {
        # do not estimate variance
        # so also if we wish to use the Rasch Model but not estimate the variance
        # and instead fix it to 1, just don't put itemType = 'Rasch'
        values[length(values)] <- 1
    }
    lbounds <- c(rep(c(-Inf,-Inf,0,0),times = ncol(x)), -Inf, 1e-04)
    ubounds <-c(rep(c(Inf,Inf,1,1),times = ncol(x)), Inf, Inf)
    tobeEstimated <- c(rep(c('alpha' %in% paramIds,
                             'delta' %in% paramIds,
                             'guessing' %in% paramIds,
                             FALSE), ncol(x)), FALSE, estimate_var)
    classes <- c(rep("dich", 4*ncol(x)), 'GroupPars', 'GroupPars')
    # The repackaged parameter data frame:
    params <- data.frame(group = "all", item = itemsAGroup, class = classes,
                         name = names,
                         parnum = as.integer(seq(from=1, to=length(names), by = 1)),
                         value = values, lbound = lbounds, ubound = ubounds,
                         est = tobeEstimated,
                         prior.type = "none", prior_1 = NaN, prior_2 = NaN,
                         stringsAsFactors = FALSE)
    return(params)
}

# IRT_F1 ---------------------------------------------------------------------------------------
# The fixed point function for case 1.
# The input p contains the current parameter estimates for the deltas.
# The last entry in p is the variance. It is also estimated.
# The input x contains the IRT data.
IRT_F1 <- function(p, x, pstar = NULL, dimx = NULL, paramIds = c('delta'),
                   estimate_var = TRUE, itemType = 'Rasch'){
    # Repackage p in a way that is understood by mirt.
    params <- repackageParams(p, x, paramIds, estimate_var)

    # Call mirt for one EM update.
    result <- mirt(x, model = 1, itemtype = itemType, method = 'EM', optimizer= 'BFGS', dentype = 'Gaussian',
                   TOL = 10e-7, technical = list(NCYCLES = 1, message=F), pars = params,
                   verbose = F, accelerate='none')

    # Return the new parameter estimates (d = delta).
    return(c(coef(result, simplify = TRUE)$items[,"d"],
             coef(result, simplify = TRUE)$cov))
}

# IRT_F2 -------------------------------------------------------------------------------------
# The fixed point function for case 2.
# The input p contains the current parameter estimates for the deltas and the alphas.
# The last entry in p is the variance. It is not estimated.
# The input x contains the IRT data.
IRT_F2 <- function(p, x, pstar = NULL, dimx = NULL, paramIds = c('delta', 'alpha'),
                   estimate_var = FALSE, itemType = '2PL'){
    # Repackage p in a way that is understood by mirt.
    params <- repackageParams(p, x, paramIds, estimate_var)

    # Call mirt for one EM update.
    result <- mirt(x, model = 1, itemtype = itemType, method = 'EM', optimizer= 'BFGS', dentype = 'Gaussian',
                   TOL = 10e-7, technical = list(NCYCLES = 1, message=F), pars = params,
                   verbose = F, accelerate='none')

    # Return the new parameter estimates (d = delta, a1 = alpha).
    return(c(coef(result, simplify = TRUE)$items[,"d"],
             coef(result, simplify = TRUE)$items[,"a1"],
             coef(result, simplify = TRUE)$cov))
}

# IRT_NLL ---------------------------------------------------------------------------------
# The negative observed log-likelihood for case 1 and case 2.
IRT_NLL <- function(p,x, paramIds, estimate_var, itemType){
    # Repackage p in a way that is understood by mirt.
    params <- repackageParams(p, x, paramIds, estimate_var)

    # From mirt.R
    Call <- match.call()
    latent.regression <- mirt:::latentRegression_obj(data=x, covdata=NULL,
                                              dentype='Gaussian', formula=NULL, method='EM')

    # From 03-estimation.R and EMstep.group.R
    NLL(data=x, model=1, group=rep('all', nrow(x)),
                      itemtype=itemType, guess=0, upper=1, grsm.block=NULL,
                      pars=params, method='EM', constrain=NULL, SE=FALSE, TOL= 10e-7,
                      parprior=NULL, quadpts=NULL, monopoly.k=1L,
                      technical=list(NCYCLES = 1, message=F), verbose=F, survey.weights=NULL,
                      calcNull=F, SE.type='Oakes', large=F, key=NULL,
                      accelerate='none', draws=5000, rsm.block=NULL,
                      GenRandomPars=F, optimizer=NULL,
                      solnp_args=list(), nloptr_args=list(),
                      latent.regression=latent.regression, gpcm_mats=list(),
                      control=list(), spline_args=list(), dentype='Gaussian')
}

# NLL -----------------------------------------------------------------------------------------
# From 03-estimation.R
NLL <- function(data, model, group, itemtype = NULL, guess = 0, upper = 1,
                invariance = '', pars = NULL, constrain = NULL, key = NULL,
                parprior = NULL, mixed.design = NULL, customItems = NULL,
                customItemsData = NULL, customGroup = NULL,
                GenRandomPars = FALSE, large = FALSE,
                survey.weights = NULL, latent.regression = NULL,
                gpcm_mats=list(), spline_args=list(), monopoly.k=1,
                control = list(), ...)
{
    start.time <- proc.time()[3L]
    dots <- list(...)
    if(missing(data)) missingMsg('data')
    if(length(unique(colnames(data))) != ncol(data))
        stop('items must have unique names in data input', call.=FALSE)
    if(missing(model)) missingMsg('model')
    if(missing(group)) missingMsg('group')
    if(!(is.factor(group) || is.character(group)) || length(group) != nrow(data))
        stop('group input provided is not valid', call.=FALSE)
    if(is.logical(large) && large){
        Data <- opts <- list()
        opts$zeroExtreme <- FALSE
        opts$dentype <- 'default'
        if(!is.null(dots$technical$zeroExtreme)) opts$zeroExtreme <- dots$technical$zeroExtreme
        data <- as.matrix(data)
        if(is.numeric(data))
            data <- matrix(as.integer(data), nrow(data), dimnames=list(rownames(data), colnames(data)))
        rownames(data) <- 1L:nrow(data)
        if(is.null(colnames(data)))
            colnames(data) <- paste0('Item.', 1L:ncol(data))
        Data$data <- data
        Data$group <- factor(group)
        Data$groupNames <- levels(Data$group)
        Data$ngroups <- length(Data$groupNames)
        Data$nitems <- ncol(data)
        K <- apply(Data$data, 2L, function(x) length(unique(na.omit(x))))
        tmp <- list(...)
        if(!is.null(tmp$technical$customK)) K <- tmp$technical$customK
        itemloc <- c(1L, cumsum(K) + 1L)
        Names <- NULL
        for(i in 1L:length(K))
            Names <- c(Names, paste("Item.",i,"_",1L:K[i],sep=""))
        PrepListFull <- list(K=K, itemloc=itemloc,
                             Names=Names, itemnames=colnames(Data$data))
    } else {
        if(missing(data) || is.null(nrow(data)))
            stop('data argument is required', call.=FALSE)
        if(missing(model) || !is(model, 'numeric') && !is(model, 'mirt.model') &&
           !is(model, 'character'))
            stop('model argument (numeric, character, or from mirt.model() function) is required',
                 call.=FALSE)
        if(!(is.factor(group) || is.character(group)) || length(group) != nrow(data))
            stop('group input provided is not valid', call.=FALSE)
        if(!is.null(itemtype))
            stopifnot(is(itemtype, 'character'))
        if(!is.null(constrain))
            stopifnot(is(constrain, 'list'))
        if(!is.null(parprior))
            stopifnot(is(parprior, 'list'))
        if(!is.null(customItems))
            stopifnot(is(customItems, 'list'))
        if(!is.null(customItemsData))
            stopifnot(is(customItemsData, 'list') || length(customItemsData) == ncol(data))
        if(!is.null(customGroup))
            stopifnot(is(customGroup, 'GroupPars'))
        stopifnot(is(invariance, 'character'))
        stopifnot(is(GenRandomPars, 'logical'))
        stopifnot(is(large, 'logical') || is(large, 'list'))
        opts <- mirt:::makeopts(GenRandomPars=GenRandomPars,
                                hasCustomGroup=!is.null(customGroup), ...)
        if(opts$Moptim == 'NR'){
            if(is.null(control$tol)) control$tol <- opts$TOL/1000
            if(is.null(control$maxit)) control$maxit <- 50L
        }
        if(opts$Moptim == 'nlminb')
            if(is.null(control$rel.tol)) control$rel.tol <- opts$TOL/100

        if(opts$dentype == "discrete" && is.null(customGroup)){
            den <- Theta_discrete_den
            par <- if(!is.null(dots$structure)){
                if(is.null(opts$technical$customTheta))
                    stop('customTheta must be defined when using structure input', call.=FALSE)
                opts$structure <- model.matrix(dots$structure, as.data.frame(opts$technical$customTheta))
                tmpnfact <- ncol(opts$technical$customTheta)
                numeric(ncol(opts$structure))
            } else if(is.null(opts$technical$customTheta)){
                tmpnfact <- model
                numeric(model-1L)
            } else {
                tmpnfact <- ncol(opts$technical$customTheta)
                numeric(nrow(opts$technical$customTheta) - 1L)
            }
            if(length(par)){
                names(par) <- paste0('c', 1:length(par))
                est <- rep(TRUE, length(par))
            } else {
                par <- c(c = 0)
                est <- FALSE
            }
            customGroup <- createGroup(par=par, est=est, den=den, nfact=tmpnfact,
                                       gen=function(object) rnorm(length(object@par), 0, 1/2))
            customGroup@itemclass <- -1L
        }
        if(!is.null(survey.weights)){
            stopifnot(opts$method %in% c('EM', 'QMCEM', 'MCEM'))
            stopifnot(length(survey.weights) == nrow(data))
        }
        if(any(is.na(group))){
            if(opts$message)
                message('NA values in group removed, along with associated rows in data')
            data <- data[!is.na(group), , drop=FALSE]
            group <- group[!is.na(group)]
        }
        if(!is.null(customItems) || any(itemtype %in% mirt:::Experimental_itemtypes()))
            opts$calcNull <- FALSE
        opts$times <- list(start.time=start.time)
        # on exit, reset the seed to override internal
        if(opts$method == 'MHRM' || opts$method == 'MIXED' || opts$method == 'SEM' &&
           opts$plausible.draws == 0L)
            on.exit(set.seed((as.numeric(Sys.time()) - floor(as.numeric(Sys.time()))) * 1e8))
        #change itemtypes if NULL.MODEL
        if(opts$NULL.MODEL){
            constrain <- NULL
            if(!is.null(itemtype)){
                itemtype[itemtype == 'grsm'] <- 'graded'
                itemtype[itemtype == 'rsm'] <- 'gpcm'
                itemtype[itemtype == '3PL' | itemtype == '3PLu' | itemtype == '4PL'] <- '2PL'
                itemtype[itemtype == '3PLNRM' | itemtype == '3PLuNRM' | itemtype == '4PLNRM'] <- '2PLNRM'
                itemtype[itemtype == 'spline'] <- '2PL'
            }
        }
        if(!is.null(itemtype)){
            if(any(itemtype == 'spline') && !(opts$method %in% c('EM', 'QMCEM', 'MCEM')))
                stop('spline itemtype only supported for EM algorithm', call.=FALSE)
        }
        if(length(group) != nrow(data))
            stop('length of group not equal to number of rows in data.', call.=FALSE)
        if(any(is.na(group)))
            stop('Unknown group memberships cannot be estimated. Please remove the NA values in group
                    and the associated rows in the data input.', call.=FALSE)

        #####
        ### uncomment when building html examples
        # opts$verbose <- FALSE
        ####
        opts$times$start.time.Data <- proc.time()[3L]
        Data <- list()
        data <- as.matrix(data)
        if(!all(apply(data, 2L, class) %in% c('integer', 'numeric')))
            stop('Input data must be integer or numeric values only', call.=FALSE)
        if(is.numeric(data))
            data <- matrix(as.integer(data), nrow(data),
                           dimnames=list(rownames(data), colnames(data)))
        rownames(data) <- 1L:nrow(data)
        if(is.null(colnames(data)))
            colnames(data) <- paste0('Item.', 1L:ncol(data))
        if(nrow(data) > 1L && is.null(opts$technical$customK)){
            if(!is.null(key)){
                key <- sapply(1L:length(key), function(ind, data, key){
                    s <- sort(unique(data[,ind]))
                    se <- min(s, na.rm = TRUE):(length(s) + min(s) - 1L)
                    ret <- se[s == key[ind]]
                    if(!length(ret)) ret <- NA
                    ret
                }, data=data, key=key)
            }
            data <- apply(data, 2L, function(x, message){
                s <- sort(unique(x))
                se <- min(s, na.rm = TRUE):max(x, na.rm = TRUE)
                if(length(s) != length(se)){
                    if(message)
                        message('Item re-scored so that all values are within a distance of 1')
                    for(i in 2L:length(s))
                        x <- ifelse(x == s[i], se[i], x)
                }
                x
            }, message = opts$message)
        }
        if(any(rowSums(is.na(data)) == ncol(data))){
            if(!opts$removeEmptyRows)
                stop('data contains completely empty response patterns. ',
                     'Please remove manually or pass removeEmptyRows=TRUE to the technical list',
                     call.=FALSE)
            else {
                pick <- rowSums(is.na(data)) != ncol(data)
                data <- subset(data, pick)
                group <- subset(group, pick)
                if(!is.null(latent.regression) || !is.null(mixed.design))
                    stop('removeEmptyRows input not supported for latent regression/mixed effect models. ',
                         'Please remove the require rows manually for each object.', call.=FALSE)
            }
        }
        Data$data <- data

        if(is.null(opts$grsm.block)) Data$grsm.block <- rep(1L, ncol(data))
        else Data$grsm.block <- opts$grsm.block
        if(is.null(opts$rsm.block)) Data$rsm.block <- rep(1L, ncol(data))
        else Data$rsm.block <- opts$rsm.block
        Data$group <- factor(group)
        Data$groupNames <- levels(Data$group)
        if(any(grepl('-', Data$groupNames)))
            stop('Group names cannot contain a dash (-) character', call.=FALSE)
        Data$ngroups <- if(!is.null(opts$ngroups)) opts$ngroups else length(Data$groupNames)
        Data$nitems <- ncol(Data$data)
        Data$N <- nrow(Data$data)
        Data$mins <- suppressWarnings(apply(data, 2L, min, na.rm=TRUE))
        Data$mins[!is.finite(Data$mins)] <- 0L
        if(is.character(model)){
            tmp <- any(sapply(colnames(data), grepl, x=model))
            model <- mirt.model(model, itemnames = if(tmp) colnames(data) else NULL)
        }
        oldmodel <- model
        PrepList <- vector('list', Data$ngroups)
        tmp <- 1L:Data$ngroups
        if(opts$dentype == "mixture")
            Data$groupNames <- paste0("MIXTURE_", seq_len(Data$ngroups))
        names(PrepList) <- Data$groupNames
        model <- mirt:::buildModelSyntax(model, J=Data$nitems, groupNames=Data$groupNames,
                                         itemtype=itemtype)
        Data$model <- model
        if(!is.null(dots$PrepList)) {
            PrepListFull <- PrepList[[1L]] <- dots$PrepList
        } else {
            PrepListFull <- PrepList[[1L]] <-
                mirt:::PrepData(data=Data$data, model=Data$model, itemtype=itemtype, guess=guess,
                                upper=upper, parprior=parprior, verbose=opts$verbose,
                                technical=opts$technical, parnumber=1L, BFACTOR=opts$dentype == 'bfactor',
                                grsm.block=Data$grsm.block, rsm.block=Data$rsm.block,
                                mixed.design=mixed.design, customItems=customItems,
                                customItemsData=customItemsData,
                                customGroup=customGroup, spline_args=spline_args, monopoly.k=monopoly.k,
                                fulldata=opts$PrepList[[1L]]$fulldata, key=key, opts=opts,
                                gpcm_mats=gpcm_mats, internal_constraints=opts$internal_constraints,
                                dcIRT_nphi=opts$dcIRT_nphi, dentype=opts$dentype, item.Q=opts$item.Q)
            if(!is.null(dots$Return_PrepList)) return(PrepListFull)
        }
        if(!is.null(opts$structure)){
            PrepList[[1L]]$pars[[Data$nitems+1L]]@structure <- opts$structure
            PrepList[[1L]]$pars[[Data$nitems+1L]]@parnames <-
                names(PrepList[[1L]]$pars[[Data$nitems+1L]]@est) <- colnames(opts$structure)
            PrepList[[1L]]$pars[[Data$nitems+1L]]@itemclass <- -999L
        }
        parnumber <- max(PrepList[[1L]]$pars[[Data$nitems+1L]]@parnum) + 1L
        attr(PrepListFull$pars, 'nclasspars') <- attr(PrepList[[1L]]$pars, 'nclasspars') <-
            sapply(PrepListFull$pars, function(y) length(y@parnum))
        for(g in seq_len(Data$ngroups)){
            if(g != 1L){
                PrepList[[g]] <- list(pars=PrepList[[1L]]$pars)
                for(i in 1L:length(PrepList[[g]]$pars)){
                    PrepList[[g]]$pars[[i]]@parnum <- parnumber:(parnumber + length(PrepList[[g]]$pars[[i]]@parnum) - 1L)
                    parnumber <- max(PrepList[[g]]$pars[[i]]@parnum) + 1L
                }
            }
        }
        if(length(mixed.design$random) > 0L){
            for(i in seq_len(length(mixed.design$random))){
                mixed.design$random[[i]]@parnum <- parnumber:(parnumber - 1L +
                                                                  length(mixed.design$random[[i]]@par))
                parnumber <- max(mixed.design$random[[i]]@parnum) + 1L
            }
        }
        if(!is.null(latent.regression)){
            if(length(PrepListFull$prodlist))
                stop('Polynomial combinations currently not supported when latent regression effects are used', call.=FALSE)
            lrPars <- make.lrdesign(df=latent.regression$df, formula=latent.regression$formula,
                                    factorNames=PrepListFull$factorNames, EM=latent.regression$EM,
                                    TOL=opts$TOL)
            lrPars@parnum <- parnumber:(parnumber - 1L + length(lrPars@par))
            parnumber <- max(lrPars@parnum) + 1L
            if(opts$dentype == 'discrete'){
                tmp <- matrix(1L:length(lrPars@beta), nrow(lrPars@beta), ncol(lrPars@beta))
                tmp2 <- tmp[1, ]
                lrPars@est[tmp2[-length(tmp2)]] <- TRUE
                lrPars@est[tmp[,ncol(tmp)]] <- FALSE
                PrepList$all$pars[[ncol(data) + 1L]]@est <- FALSE
            }
        } else lrPars <- list()
        if(length(latent.regression$lr.random) > 0L){
            for(i in seq_len(length(latent.regression$lr.random))){
                latent.regression$lr.random[[i]]@parnum <- parnumber:(parnumber - 1L +
                                                                          length(latent.regression$lr.random[[i]]@par))
                parnumber <- max(latent.regression$lr.random[[i]]@parnum) + 1L
            }
        }
    }
    if(!is.null(opts$PrepList)){
        PrepList <- opts$PrepList
    } else {
        if(!is.list(large)){
            tmpdata <- Data$data
            if(opts$zeroExtreme){
                n_is_na <- rowSums(is.na(tmpdata))
                sums <- colSums(t(tmpdata) - Data$mins, na.rm=TRUE)
                maxs <- apply(tmpdata, 2L, function(x) length(unique(na.omit(x)))) - 1L
                if(is.null(survey.weights)) survey.weights <- rep(1, nrow(tmpdata))
                survey.weights[sums == 0L | (sums + n_is_na) == sum(maxs)] <- 0
            }
            tmpdata[is.na(tmpdata)] <- 99999L
            stringfulldata <- apply(tmpdata, 1L, paste, sep='', collapse = '/')
            stringtabdata <- unique(stringfulldata)
            tmptabdata <- mirt:::maketabData(stringfulldata=stringfulldata, stringtabdata=stringtabdata,
                                             group=Data$group,
                                             groupNames=if(opts$dentype != 'mixture') Data$groupNames else 'full',
                                             nitem=Data$nitems,
                                             K=PrepListFull$K, itemloc=PrepListFull$itemloc,
                                             Names=PrepListFull$Names, itemnames=PrepListFull$itemnames,
                                             survey.weights=survey.weights)
            if(is.logical(large) && large){
                return(tmptabdata)
            } else large <- tmptabdata
        }
        Data$tabdatalong <- large$tabdata
        Data$tabdata <- large$tabdata2
        for(g in seq_len(Data$ngroups)){
            if(opts$dentype == 'mixture'){
                Data$fulldata[[g]] <- PrepListFull$fulldata
                Data$Freq[[g]] <- large$Freq[[1L]]
            } else {
                select <- Data$group == Data$groupNames[g]
                Data$fulldata[[g]] <- PrepListFull$fulldata[select, , drop=FALSE]
                Data$Freq[[g]] <- large$Freq[[g]]
            }
        }
    }
    if(opts$returnPrepList) return(PrepList)

    PrepList <- mirt:::UpdateParameters(PrepList, model, groupNames=Data$groupNames)
    if(GenRandomPars){
        for(g in seq_len(Data$ngroups))
            for(i in seq_len(length(PrepList[[g]]$pars)))
                PrepList[[g]]$pars[[i]] <- GenRandomPars(PrepList[[g]]$pars[[i]])

    }
    RETURNVALUES <- FALSE
    if(!is.null(pars)){
        if(is(pars, 'data.frame')){
            PrepList <- mirt:::UpdatePrepList(PrepList, pars, random=mixed.design$random,
                                              lrPars=lrPars, lr.random=latent.regression$lr.random,
                                              MG = TRUE)
            mixed.design$random <- attr(PrepList, 'random')
            if(any(pars$class == 'lrPars')) lrPars <- update.lrPars(pars, lrPars)
            attr(PrepList, 'random') <- NULL
        }
        if(!is.null(attr(pars, 'values')) || (is.character(pars) && pars == 'values'))
            RETURNVALUES <- TRUE
    }
    pars <- vector('list', Data$ngroups)
    for(g in seq_len(Data$ngroups))
        pars[[g]] <- PrepList[[g]]$pars
    nitems <- Data$nitems
    Data$K <- PrepList[[1L]]$K
    nfact <- PrepList[[1L]]$pars[[nitems+1L]]@nfact
    if(nfact != 1L && any(c('Rasch') %in% itemtype ) && PrepList[[1L]]$exploratory)
        stop('Rasch itemtype is for confirmatory models only.', call.=FALSE)
    nLambdas <- PrepList[[1L]]$pars[[1L]]@nfact
    if(is.null(constrain)) constrain <- list()
    nspec <- ifelse(!is.null(attr(model, 'nspec')), attr(model, 'nspec'), 1L)
    #default MG uses configural model (independent groups but each identified)
    if('free_means' %in% invariance ){ #Free factor means (means 0 for ref)
        if(opts$dentype == 'bfactor'){
            for(g in 2L:Data$ngroups)
                pars[[g]][[nitems + 1L]]@est[1L:(nfact-nspec)] <- TRUE
        } else {
            for(g in 2L:Data$ngroups)
                pars[[g]][[nitems + 1L]]@est[1L:nfact] <- TRUE
        }
    }
    dummymat <- matrix(FALSE, pars[[1L]][[nitems + 1L]]@nfact, pars[[1L]][[nitems + 1L]]@nfact)
    if(any('free_var' %in% invariance)){ #Free factor vars (vars 1 for ref)
        if(opts$dentype == 'bfactor'){
            tmp <- dummymat[1L:(nfact-nspec),1L:(nfact-nspec), drop=FALSE]
            diag(tmp) <- TRUE
            dummymat[1L:(nfact-nspec),1L:(nfact-nspec)] <- tmp
        } else {
            diag(dummymat) <- TRUE
        }
        tmp <- dummymat[lower.tri(dummymat, TRUE)]
        if(opts$dentype %in% c('Davidian', 'EHW')){
            for(g in 2L:Data$ngroups)
                pars[[g]][[nitems + 1L]]@est[2L] <- TRUE
        } else {
            for(g in 2L:Data$ngroups){
                pars[[g]][[nitems + 1L]]@est <- c(pars[[g]][[nitems + 1L]]@est[1L:pars[[g]][[nitems + 1L]]@nfact], tmp)
                names(pars[[g]][[nitems + 1L]]@est) <- names(pars[[g]][[nitems + 1L]]@par)
                pars[[g]][[nitems + 1L]]@parnames <- names(pars[[g]][[nitems + 1L]]@est)
            }
        }
    }
    if(opts$dentype == 'mixture'){
        tmp <- length(pars[[1L]][[nitems + 1L]]@par)
        pars[[1L]][[nitems + 1L]]@est[tmp] <- FALSE
        for(g in 1L:Data$ngroups)
            pars[[g]][[nitems + 1L]]@par[tmp] <- g - 1
        names(PrepList) <- Data$groupNames
    }
    if(RETURNVALUES){
        for(g in seq_len(Data$ngroups))
            PrepList[[g]]$pars <- pars[[g]]
        return(ReturnPars(PrepList, PrepList[[1L]]$itemnames, lr.random=latent.regression$lr.random,
                          random=mixed.design$random, lrPars=lrPars, MG = TRUE))
    }
    constrain <- mirt:::UpdateConstrain(pars=pars, constrain=constrain, invariance=invariance, nfact=Data$nfact,
                                        nLambdas=nLambdas, J=nitems, ngroups=Data$ngroups, PrepList=PrepList,
                                        method=opts$method, itemnames=PrepList[[1L]]$itemnames, model=model,
                                        groupNames=Data$groupNames)
    startlongpars <- c()
    if(opts$NULL.MODEL){
        constrain <- list()
        for(g in seq_len(length(pars))){
            for(i in seq_len(nitems)){
                pars[[g]][[i]] <- set_null_model(pars[[g]][[i]])
            }
        }
    }
    rr <- 0L
    for(g in seq_len(Data$ngroups)){
        if(g > 1L && opts$dentype == 'mixture') break
        r <- Data$Freq[[g]]
        rr <- rr + r
    }
    df <- prod(PrepList[[1L]]$K) - 1
    if(df > 1e10) df <- 1e10
    nestpars <- nconstr <- 0L
    for(g in seq_len(Data$ngroups))
        for(i in seq_len(nitems+1L))
            nestpars <- nestpars + sum(pars[[g]][[i]]@est)
    if(!is.null(mixed.design$random)){
        for(i in seq_len(length(mixed.design$random)))
            nestpars <- nestpars + sum(mixed.design$random[[i]]@est)
    }
    if(length(lrPars))
        nestpars <- nestpars + sum(lrPars@est)
    if(!is.null(dots$structure)) nestpars <- nestpars - 1L
    if(length(constrain) > 0L)
        for(i in seq_len(length(constrain)))
            nconstr <- nconstr + length(constrain[[i]]) - 1L
    if(Data$ngroups > 1L && !length(constrain)){
        if(opts$warn && any(invariance %in% c('free_means', 'free_var')))
            warning('Multiple-group model may not be identified without providing anchor items',
                    call.=FALSE)
        for(j in seq_len(Data$ngroups)){
            if(opts$dentype != 'mixture')
                tmp <- apply(subset(Data$data, Data$group == Data$groupNames[j]), 2L,
                             function(x) length(unique(na.omit(x)))) == Data$K
            for(i in which(!tmp)){
                if(any(PrepList[[j]]$pars[[i]]@est))
                    stop(paste0('Multiple Group model will not be identified without ',
                                'proper constraints (groups contain missing data patterns ',
                                'where item responses have been completely omitted or, alternatively, ',
                                'the number of categories within each group is not equal to the ',
                                'total number of categories)'),
                         call. = FALSE)
            }
        }
    }
    nmissingtabdata <- sum(is.na(rowSums(Data$tabdata)))
    dfsubtr <- nestpars - nconstr
    if(opts$dentype == 'EH') dfsubtr <- dfsubtr + (opts$quadpts - 1L) * Data$ngroups
    else if(opts$dentype == 'EHW') dfsubtr <- dfsubtr + (opts$quadpts - 3L) * Data$ngroups
    if(df <= dfsubtr)
        stop('Too few degrees of freedom. There are only ', df, ' degrees of freedom but ',
             dfsubtr, ' parameters were freely estimated.', call.=FALSE)
    df <- df - dfsubtr
    if(!is.null(customItems)){
        for(g in seq_len(Data$ngroups))
            PrepList[[g]]$exploratory <- FALSE
    }
    G2group <- numeric(Data$ngroups)
    DERIV <- vector('list', Data$ngroups)
    for(g in seq_len(Data$ngroups)){
        DERIV[[g]] <- vector('list', Data$nitems)
        for(i in seq_len(Data$nitems))
            DERIV[[g]][[i]] <- getMethod(mirt:::Deriv, c(class(pars[[g]][[i]]), 'matrix'))
    }
    Ls <- mirt:::makeLmats(pars, constrain, random = mixed.design$random,
                           lr.random=latent.regression$lr.random, lrPars=lrPars)
    CUSTOM.IND <- which(sapply(pars[[1L]], class) %in% mirt:::Use_R_ProbTrace())
    SLOW.IND <- which(sapply(pars[[1L]], class) %in% mirt:::Use_R_Deriv())
    if(pars[[1]][[length(pars[[1L]])]]@itemclass %in% c(-1L, -999L))
        SLOW.IND <- c(SLOW.IND, length(pars[[1L]]))
    if(opts$dentype != 'Gaussian' && opts$method %in% c('MHRM', 'MIXED', 'SEM'))
        stop('Non-Gaussian densities not currently supported with MHRM algorithm')
    #warnings
    wmsg <- 'Lower and upper bound parameters (g and u) should use \'norm\' (i.e., logit) prior'
    for(g in seq_len(length(pars))){
        for(i in seq_len(length(pars[[1L]]))){
            if(class(pars[[g]][[i]]) == 'dich'){
                pt <- pars[[g]][[i]]@prior.type
                if(!(pt[length(pt)-1L] %in% c(0L, 1L, 4L))) warning(wmsg, call.=FALSE)
                if(!(pt[length(pt)] %in% c(0L, 1L, 4L))) warning(wmsg, call.=FALSE)
                next
            } else if(class(pars[[g]][[i]]) == 'partcomp'){
                pt <- pars[[g]][[i]]@prior.type
                if(!(pt[length(pt)-1L] %in% c(0L, 1L))) warning(wmsg, call.=FALSE)
                if(!(pt[length(pt)] %in% c(0L, 1L))) warning(wmsg, call.=FALSE)
                next
            } else if(class(pars[[g]][[i]]) == 'nestlogit'){
                pt <- pars[[g]][[i]]@prior.type
                if(!(pt[nfact + 2L] %in% c(0L, 1L))) warning(wmsg, call.=FALSE)
                if(!(pt[nfact + 3L] %in% c(0L, 1L))) warning(wmsg, call.=FALSE)
                next
            }
        }
    }
    SEMconv <- NA
    opts$times$end.time.Data <- proc.time()[3L]

    # Get negative log-likelihood from EStep
    opts$times$start.time.Estimate <- proc.time()[3L]
    if(length(lrPars)){
        if(opts$SE && !(opts$SE.type %in% c('complete', 'forward', 'central', 'Richardson')))
            stop('Information matrix method for latent regression estimates not supported',
                 call.=FALSE)
        opts$full <- TRUE
    } else opts$full <- FALSE
    temp <- matrix(0L,nrow=nitems,ncol=nspec)
    sitems <- matrix(0L, nrow=sum(PrepList[[1L]]$K), ncol=nspec)
    specific <- NULL

    if(is.null(opts$quadpts)){
        tmp <- if(opts$dentype == 'bfactor') PrepList[[1L]]$nfact - attr(model, 'nspec') + 1L
        else nfact
        opts$quadpts <- mirt:::select_quadpts(tmp)
    }
    if(opts$quadpts < 3 && opts$warn) warning('Should use more than 2 quadpts', call.=FALSE)
    theta <- 1
    if(!(opts$method %in% c('QMCEM', 'MCEM')))
        theta <- as.matrix(seq(opts$theta_lim[1L], opts$theta_lim[2L],
                               length.out = opts$quadpts))
    if(opts$quadpts^nfact <= opts$MAXQUAD){
        if(is.null(opts$technical$customTheta))
            Theta <- thetaComb(theta, nfact)
    } else stop('Greater than ', opts$MAXQUAD, ' quadrature points.', call.=FALSE)
    if(opts$message && nfact > 3L && !(opts$odentype %in% c('custom', 'discrete')))
        message('EM quadrature for high dimensional models are better handled
                                 \twith the \"QMCEM\" or \"MCEM\" method')


    if(!is.null(opts$technical$customTheta)){
        Theta <- opts$technical$customTheta
        if(!is.matrix(Theta)) stop('customTheta input must be a matrix', call.=FALSE)
        opts$quadpts <- nrow(Theta)
    }
    pars <- mirt:::loadSplinePars(pars, Theta)
    #end Theta def
    NLL <- NLL_FROM_ESTEP(pars=pars, constrain=constrain, Ls=Ls, PrepList=PrepList, Data=Data,
                          list = list(NCYCLES=opts$NCYCLES, TOL=opts$TOL, MSTEPTOL=opts$MSTEPTOL,
                                      nfactNames=PrepList[[1L]]$nfactNames, theta=theta,
                                      itemloc=PrepList[[1L]]$itemloc, dentype=opts$dentype,
                                      sitems=sitems, specific=specific, NULL.MODEL=opts$NULL.MODEL,
                                      nfact=nfact, constrain=constrain, verbose=opts$verbose,
                                      SE = opts$SE, SE.type=opts$SE.type, delta=opts$delta, quadpts=opts$quadpts,
                                      accelerate=opts$accelerate, CUSTOM.IND=CUSTOM.IND, SLOW.IND=SLOW.IND,
                                      customPriorFun=opts$customPriorFun, Moptim=opts$Moptim, warn=opts$warn,
                                      message=opts$message, method=opts$method, full=opts$full,
                                      lrPars=lrPars, SE=opts$SE && opts$SE.type == 'numerical', Etable=opts$Etable,
                                      NULL.MODEL=opts$NULL.MODEL, PLCI=opts$PLCI, Norder=opts$Norder,
                                      keep_vcov_PD=opts$keep_vcov_PD, symmetric=opts$technical$symmetric,
                                      MCEM_draws=opts$MCEM_draws),
                          Theta=Theta, DERIV=DERIV, solnp_args=opts$solnp_args, control=control)
    return(NLL)
}

# NLL_FROM_ESTEP ------------------------------------------------------------------------------------
# From EMstep.group.R
NLL_FROM_ESTEP <- function(pars, constrain, Ls, Data, PrepList, list, Theta, DERIV, solnp_args, control)
{
    verbose <- list$verbose
    lrPars <- list$lrPars
    nfact <- list$nfact
    if(list$dentype %in% c('EH', 'EHW') && nfact != 1L)
        stop('empirical histogram only available for unidimensional models', call.=FALSE)
    if(list$dentype == 'Davidian'){
        if(nfact != 1L)
            stop('Davidian curves estimation only available for unidimensional models', call.=FALSE)
        if(list$SE)
            stop('No standard error method currently supported for Davidian curves', call.=FALSE)
        J <- length(pars[[1L]]) - 1L
        for(g in seq_len(length(pars))){ # throw error if latent mean/var estimated TODO
            if(any(pars[[g]][[J+1L]]@est[1L:2L]))
                stop('Estimating Davidian mean-variance hyper-parameters is not currently supported',
                     call.=FALSE)
        }
    }
    NCYCLES <- list$NCYCLES
    TOL <- list$TOL
    CUSTOM.IND <- list$CUSTOM.IND
    dentype <- list$dentype
    itemloc <- list$itemloc
    ngroups <- length(pars)
    specific <- list$specific
    sitems <- list$sitems
    theta <- list$theta
    full <- list$full
    J <- length(itemloc) - 1L
    nfullpars <- 0L
    estpars <- c()
    prodlist <- PrepList[[1L]]$prodlist
    MC <- list$method %in% c('QMCEM', 'MCEM')
    QMC <- list$method == 'QMCEM'
    for(g in seq_len(ngroups)){
        for(i in seq_len(J+1L)){
            nfullpars <- nfullpars + length(pars[[g]][[i]]@par)
            estpars <- c(estpars, pars[[g]][[i]]@est)
        }
        if(length(lrPars)){
            nfullpars <- nfullpars + length(lrPars@par)
            estpars <- c(estpars, lrPars@est)
        }
    }
    listpars <- vector('list', ngroups)
    for(g in seq_len(ngroups)){
        listpars[[g]] <- list()
        for(i in seq_len(J + 1L)){
            listpars[[g]][[i]] <- pars[[g]][[i]]@par
        }
        if(length(lrPars))
            listpars[[g]][[i+1L]] <- lrPars@par
    }
    index <- seq_len(nfullpars)
    longpars <- rep(NA,nfullpars)
    latent_longpars <- logical(nfullpars)
    ind1 <- 1L
    for(g in seq_len(ngroups)){
        for(i in seq_len(J+1L)){
            ind2 <- ind1 + length(pars[[g]][[i]]@par) - 1L
            longpars[ind1:ind2] <- pars[[g]][[i]]@par
            if(i == (J+1L)) latent_longpars[ind1:ind2] <- TRUE
            ind1 <- ind2 + 1L
        }
        if(length(lrPars)){
            ind2 <- ind1 + length(lrPars@par) - 1L
            longpars[ind1:ind2] <- lrPars@par
            ind1 <- ind2 + 1L
        }
    }
    converge <- TRUE
    L <- Ls$L
    redun_constr <- Ls$redun_constr
    estindex_unique <- index[estpars & !redun_constr]
    if(any(diag(L)[!estpars] > 0L) && !list$PLCI){
        redindex <- index[!estpars]
        stop('Constraint applied to fixed parameter(s) ',
             paste(paste0(redindex[diag(L)[!estpars] > 0L]), ''), ' but should only be applied to
                 estimated parameters. Please fix!', call.=FALSE)
    }
    prior <- rlist <- r <- vector('list', ngroups)
    #make sure constrained pars are equal
    tmp <- rowSums(L)
    tmp[tmp == 0L] <- 1L
    check <- as.numeric(L %*% longpars) / tmp
    longpars[estpars] <- check[estpars]
    LL <- LP <- 0
    LBOUND <- UBOUND <- c()
    for(g in seq_len(ngroups)){
        for(i in seq_len(J+1L)){
            LBOUND <- c(LBOUND, pars[[g]][[i]]@lbound)
            UBOUND <- c(UBOUND, pars[[g]][[i]]@ubound)
        }
        if(length(lrPars)){
            LBOUND <- c(LBOUND, lrPars@lbound)
            UBOUND <- c(UBOUND, lrPars@ubound)
        }
    }
    est <- c()
    for(g in seq_len(ngroups)){
        for(j in seq_len(J+1L))
            est <- c(est, pars[[g]][[j]]@est)
        if(length(lrPars))
            est <- c(est, rep(FALSE, length(lrPars@est)))
    }
    for(i in seq_len(length(constrain)))
        est[constrain[[i]][-1L]] <- FALSE
    if(list$dentype %in% c('Davidian', 'EHW')) # dont estimate mean/var for extrapolation
        est[names(est) %in% c('MEAN_1', 'COV_11')] <- FALSE
    if(all(!est) && list$SE)
        stop('Computing ACOV matrix is meaningless when no parameters are estimated', call.=FALSE)
    names(longpars) <- names(est)
    if(list$Moptim != 'BFGS') {
        Moptim <- list$Moptim
    } else {
        Moptim <- if(all(c(LBOUND[est], UBOUND[est]) %in% c(-Inf, Inf))) 'BFGS' else 'nlminb'
    }
    if(Moptim == 'NR' && sum(est) > 300L && list$message)
        message('NR optimizer should not be used for models with a large number of parameters.
                Use the optimizer = \'BFGS\' or \'nlminb\' instead.')
    EMhistory <- matrix(NA, NCYCLES+1L, length(longpars))
    EMhistory[1L,] <- longpars
    ANY.PRIOR <- rep(FALSE, ngroups)
    if(length(prodlist))
        Theta <- prodterms(Theta, prodlist)
    gTheta <- vector('list', ngroups)
    for(g in seq_len(ngroups)){
        ANY.PRIOR[g] <- any(sapply(pars[[g]], function(x) x@any.prior))
        gTheta[[g]] <- Theta
    }
    preMstep.longpars2 <- preMstep.longpars <- longpars
    is_SEM <- list$SE.type == 'SEM'
    accel <- 0; Mrate <- ifelse(is_SEM, 1, .4)
    Estep.time <- Mstep.time <- 0
    collectLL <- rep(NA, NCYCLES)
    hess <- matrix(0)
    Elist <- list()
    startMrate <- ifelse(Moptim == 'L-BFGS-B', 5L, 1L)
    # GET LOG-LIKELIHOOD from EStep
    #priors
    if(length(lrPars)) lrPars@mus <- lrPars@X %*% lrPars@beta
    if(MC)
        gTheta <- mirt:::updateTheta(npts=if(QMC) list$quadpts else list$MCEM_draws(cycles),
                                     nfact=nfact, pars=pars, QMC=QMC)
    tmp <- mirt:::updatePrior(pars=pars, gTheta=gTheta,
                              list=list, ngroups=ngroups, nfact=nfact,
                              J=J, dentype=dentype, sitems=sitems, cycles=cycles,
                              rlist=rlist, full=full, lrPars=lrPars, MC=MC)
    prior <- tmp$prior; Prior <- tmp$Prior; Priorbetween <- tmp$Priorbetween
    if(is.na(TOL) && !is.nan(TOL)){
        for(g in seq_len(ngroups)) rlist[[g]]$expected <- 1
        break
    }
    Elist <- mirt:::Estep(pars=pars, Data=Data, gTheta=gTheta, prior=prior, Prior=Prior,
                          Priorbetween=Priorbetween, specific=specific, sitems=sitems,
                          ngroups=ngroups, itemloc=itemloc, CUSTOM.IND=CUSTOM.IND,
                          dentype=dentype, rlist=rlist, full=full, Etable=list$Etable)
    rlist <- Elist$rlist; LL <- Elist$LL
    -LL
}

# pconstr_var -----------------------------------------------------------------------------
# used in turboem to constrain parameter space of variance of latent factor
pconstr_var <- function(par) {
  lower <- c(-Inf, 0)
  upper <- c(Inf, Inf)
  return(all(lower <= par & par < upper))
}

# generate_item_params --------------------------------------------------------------------
# model argument must be "1PL" or "2PL"
generate_item_params <- function(n_items, model) {
  if (model == "1PL" | model == "1PL_fixed_var") {
    bounds_diff <- c(-2, 2)
    mu_diff <- 0
    var_diff <- 1
    item_params <- numeric(n_items)
    for (item in 1:n_items) {
      done <- FALSE
      while (!done) {
        item_params[item] <- rnorm(1, mu_diff, var_diff)
        done <- (item_params[item] >= bounds_diff[1]) & (item_params[item] <= bounds_diff[2])
      }
    }
    item_params <- cbind(item_params, rep(1, n_items))
  } else {
    bounds_diff <- c(-2, 2)
    bounds_disc <- c(0.5, 2)
    mus <- c(0, 1)
    cor <- 0.35
    vars <- c(1, 1)
    covar <- cor * sqrt(vars[1]) * sqrt(vars[2])
    cov_matrix <- matrix(c(vars[1], covar, covar, vars[2]), 2, 2)
    item_params <- matrix(nrow = n_items, ncol = 2)
    for (item in 1:n_items) {
      done <- FALSE
      while (!done) {
        item_params[item, ] <- rmvnorm(1, mus, cov_matrix)
        done <- ((item_params[item, 1] >= bounds_diff[1]) & (item_params[item, 1] <= bounds_diff[2])) &
          ((item_params[item, 2] >= bounds_disc[1]) & (item_params[item, 2] <= bounds_disc[2]))
      }
    }
  }
  colnames(item_params) <- c("difficulties", "discriminations")
  return(item_params)
}

# simulate_data ----------------------------------------------------------------------------------
# takes vectors with item difficulties, discriminations and person abilities (all of
# class numeric)
simulate_data <- function(difficulties, discriminations, abilities) {
  # set up structure to return
  out <- list()

  # get dimensions
  n_items <- length(difficulties)
  n_persons <- length(abilities)

  # compute slopes and intercepts to have correct parameterization for mirt
  slopes <- discriminations
  intercepts <- (-1) * slopes * difficulties
  out[["slopes"]] <- slopes
  out[["intercepts"]] <- intercepts

  # for each item, compute solution probabilties,
  # so compute model equation and plug into logit function
  # and then plug into bernoulli distribution to get responses (0 or 1)
  data_sim <- data.frame(
    id = 1:n_persons
  )
  for (item in 1:n_items) {
    probs <- plogis(slopes[item]*abilities + intercepts[item])
    data_sim[[paste0("item", item)]] <- rbinom(n_persons, 1, probs)
  }
  out[["data_sim"]] <- data_sim[,-1]

  # return output
  return(out)
}

# run_simulation_trial -----------------------------------------------------------------------
# model must be either "1PL" or "1PL_fixed_var" or "2PL"
# start_values must be either "mirt" for start values used in mirt or "artifical" for less
# well-suited start values, i.e., all discriminations have start value 1 and all difficulties
# have star value 0
run_simulation_trial <- function(n_items, model, start_values) {
  library(turboEM)
  library(mirt)
  library(mvtnorm)

  n_persons <- 10000
  results <- list()

  # generate item parameters
  item_params <- generate_item_params(n_items, model)

  # generate person abilities
  abilities <- rnorm(n_persons)

  # generate data
  d <- simulate_data(item_params[, "difficulties"],
                            item_params[, "discriminations"],
                            abilities)
  data_sim <- results[["data_sim"]] <- d[["data_sim"]]

  # save true item parameters (slopes and intercepts)
  results[["true_params"]] <- cbind(d[["slopes"]], d[["intercepts"]])
  colnames(results[["true_params"]]) <- c("slopes", "intercepts")

  if (start_values == "mirt") {
    # in this case, start values as provided by mirt should be used
    # fit model with accelerator
    if (model == "1PL") {
      # fit 1PL model WITH estimating the variance

      # get initial values for deltas from mirt
      init_values_table <- mirt(data_sim, 1, 'Rasch', pars = 'values')
      initial_deltas <- init_values_table$value[init_values_table$name == "d"]
      initial_var <- init_values_table$value[init_values_table$name == "COV_11"]
      initial_values <- c(initial_deltas, initial_var)
      results[["initial_values"]] <- initial_values
      names(results[["initial_values"]]) <- c(rep("initial_deltas", n_items), "initial_var")

      # fit model with different accelerators
      results[["accelerators"]] <- turboem(par = initial_values,
                                           x = data_sim,
                                           fixptfn=IRT_F1,
                                           objfn=IRT_NLL,
                                           paramIds=c('delta'),
                                           estimate_var = TRUE,
                                           itemType = 'Rasch',
                                           method=c("em", "qn", "qn", "qn", "qn",
                                                    "squarem", "squarem", "squarem", "squarem", "pem"),
                                           control.run=list(tol=1.0e-7, convtype="parameter", maxiter = 1000),
                                           control.method=list(list(),
                                                               list(qn=1), list(qn=2), list(qn=3),list(qn=4),
                                                               list(K=1), list(K=2), list(K=3),list(K=4),
                                                               list(version="geometric", l = 10)),
                                           pconstr = pconstr_var)

    } else {
      if (model == "1PL_fixed_var") {
        # 1PL model WITHOUT estimating the variance

        # get initial values for deltas from mirt
        init_values_table <- mirt(data_sim, 1, 'Rasch', pars = 'values')
        initial_deltas <- init_values_table$value[init_values_table$name == "d"]
        initial_var <- init_values_table$value[init_values_table$name == "COV_11"]
        initial_values <- c(initial_deltas, initial_var)
        results[["initial_values"]] <- initial_values
        names(results[["initial_values"]]) <- c(rep("initial_deltas", n_items), "initial_var")

        # fit model with different accelerators
        results[["accelerators"]] <- turboem(par = initial_values,
                                             x = data_sim,
                                             fixptfn=IRT_F1,
                                             objfn=IRT_NLL,
                                             paramIds=c('delta'),
                                             estimate_var = FALSE,
                                             itemType = 'Rasch',
                                             method=c("em", "qn", "qn", "qn", "qn",
                                                      "squarem", "squarem", "squarem", "squarem", "pem"),
                                             control.run=list(tol=1.0e-7, convtype="parameter", maxiter = 1000),
                                             control.method=list(list(),
                                                                 list(qn=1), list(qn=2), list(qn=3),list(qn=4),
                                                                 list(K=1), list(K=2), list(K=3),list(K=4),
                                                                 list(version="geometric", l = 10)))
      } else {
        # 2PL model

        # get initial values for deltas and alphas from mirt
        init_values_table <- mirt(data_sim, 1, '2PL', pars = 'values')
        initial_deltas <- init_values_table$value[init_values_table$name == "d"]
        initial_alphas <- init_values_table$value[init_values_table$name == "a1"]
        initial_var <- init_values_table$value[init_values_table$name == "COV_11"]
        initial_values <- c(initial_deltas, initial_alphas, initial_var)
        results[["initial_values"]] <- initial_values
        names(results[["initial_values"]]) <- c(rep("initial_deltas", n_items),
                                                rep("initial_alphas", n_items),
                                                "initial_var")

        # fit model with different accelerators
        results[["accelerators"]] <- turboem(par= initial_values,
                                             x = data_sim,
                                             fixptfn = IRT_F2,
                                             objfn = IRT_NLL,
                                             paramIds = c('alpha','delta'),
                                             estimate_var = FALSE,
                                             itemType = '2PL',
                                             method = c("em", "qn", "qn", "qn", "qn",
                                                        "squarem", "squarem", "squarem", "squarem", "pem"),
                                             control.run = list(tol=1.0e-7, convtype="parameter", maxiter = 1000),
                                             control.method = list(list(),
                                                                   list(qn=1), list(qn=2), list(qn=3), list(qn=4),
                                                                   list(K=1), list(K=2), list(K=3), list(K=4),
                                                                   list(version="geometric", l = 10)))
        # sign switch for positive discirminations
        results[["accelerators"]]$pars[,(n_items+1):(2*n_items)] <-
          abs(results[["accelerators"]]$pars[,(n_items+1):(2*n_items)])
      }
    }
  } else {
    # in this case, start values should be set to artifical values
    # fit model with accelerator
    if (model == "1PL") {
      # fit 1PL model WITH estimating the variance

      # set initial values to artifical values
      initial_values <- c(rep(0, n_items), 1)
      results[["initial_values"]] <- initial_values
      names(results[["initial_values"]]) <- c(rep("initial_deltas", n_items), "initial_var")

      # fit model with different accelerators
      results[["accelerators"]] <- turboem(par = initial_values,
                                           x = data_sim,
                                           fixptfn=IRT_F1,
                                           objfn=IRT_NLL,
                                           paramIds=c('delta'),
                                           estimate_var = TRUE,
                                           itemType = 'Rasch',
                                           method=c("em", "qn", "qn", "qn", "qn",
                                                    "squarem", "squarem", "squarem", "squarem", "pem"),
                                           control.run=list(tol=1.0e-7, convtype="parameter", maxiter = 1000),
                                           control.method=list(list(),
                                                               list(qn=1), list(qn=2), list(qn=3),list(qn=4),
                                                               list(K=1), list(K=2), list(K=3),list(K=4),
                                                               list(version="geometric", l = 10)),
                                           pconstr = pconstr_var)

    } else {
      if (model == "1PL_fixed_var") {
        # 1PL model WITHOUT estimating the variance

        # set initial values to artififcal values
        initial_values <- c(rep(0, n_items), 1)
        results[["initial_values"]] <- initial_values
        names(results[["initial_values"]]) <- c(rep("initial_deltas", n_items), "initial_var")

        # fit model with different accelerators
        results[["accelerators"]] <- turboem(par = initial_values,
                                             x = data_sim,
                                             fixptfn=IRT_F1,
                                             objfn=IRT_NLL,
                                             paramIds=c('delta'),
                                             estimate_var = FALSE,
                                             itemType = 'Rasch',
                                             method=c("em", "qn", "qn", "qn", "qn",
                                                      "squarem", "squarem", "squarem", "squarem", "pem"),
                                             control.run=list(tol=1.0e-7, convtype="parameter", maxiter = 1000),
                                             control.method=list(list(),
                                                                 list(qn=1), list(qn=2), list(qn=3),list(qn=4),
                                                                 list(K=1), list(K=2), list(K=3),list(K=4),
                                                                 list(version="geometric", l = 10)))
      } else {
        # 2PL model

        # set initial values to artififcal values
        initial_values <- c(rep(0, n_items), rep(1, n_items), 1)
        results[["initial_values"]] <- initial_values
        names(results[["initial_values"]]) <- c(rep("initial_deltas", n_items),
                                                rep("initial_alphas", n_items),
                                                "initial_var")

        # fit model with different accelerators
        results[["accelerators"]] <- turboem(par= initial_values,
                                             x = data_sim,
                                             fixptfn = IRT_F2,
                                             objfn = IRT_NLL,
                                             paramIds = c('alpha','delta'),
                                             estimate_var = FALSE,
                                             itemType = '2PL',
                                             method = c("em", "qn", "qn", "qn", "qn",
                                                        "squarem", "squarem", "squarem", "squarem", "pem"),
                                             control.run = list(tol=1.0e-7, convtype="parameter", maxiter = 1000),
                                             control.method = list(list(),
                                                                   list(qn=1), list(qn=2), list(qn=3), list(qn=4),
                                                                   list(K=1), list(K=2), list(K=3), list(K=4),
                                                                   list(version="geometric", l = 10)))
        # sign switch for positive discirminations
        results[["accelerators"]]$pars[,(n_items+1):(2*n_items)] <-
          abs(results[["accelerators"]]$pars[,(n_items+1):(2*n_items)])
      }
    }
  }
  return(results)
}

# run simulation ------------------------------------------------------------------------------------

sim_conditions <- expand.grid(
  model = c("1PL", "1PL_fixed_var", "2PL"),
  n_items = c(9, 100),
  stringsAsFactors = FALSE
)
sim_conditions$condition <- paste0(sim_conditions$model, "_nitems", sim_conditions$n_items)

n_trials <- 100

# set.seed(123)

# set up parallel computing on cluster ---------------------------------------------------------
n_cluster <- 20
cluster <- makeCluster(n_cluster)
registerDoParallel(cluster)
registerDoRNG(123)

saved_session_info <- sessionInfo()

for (condition in 1:length(sim_conditions$model)) {
  print(condition)

  # run with mirt start values
  cond_results <- foreach(trial = 1:n_trials) %dopar%
    run_simulation_trial(n_items = sim_conditions$n_items[condition],
                         model = sim_conditions$model[condition],
                         start_values = "mirt")
  # save results
  saveRDS(cond_results, file = paste0("results_", condition, ".rds"))

  # run with artifical start values
  cond_results_art <- foreach(trial = 1:n_trials) %dopar%
    run_simulation_trial(n_items = sim_conditions$n_items[condition],
                         model = sim_conditions$model[condition],
                         start_values = "artifical")

  # save results
  saveRDS(cond_results_art, file = paste0("art_results_", condition, ".rds"))
}

# save session info
saveRDS(saved_session_info, file = "saved_session_info.rds")

# old code (commented out) --------------------------------------------------------------------------

# #################
# # IRT Case 1: F #
# #################
#
# meshSize = 4
# # just gives how many different initial values we will draw (which will be 16 in all
# # the simulations from the Bachelor's thesis)
# # and then in the Bachelor's thesis, the same initial value for delta (e.g.) was then used
# # for all items
#
# # Case 1
# # Initial delta estimates (for different trajectories).
# deltaInits <- seq(from = -3, to = 3, length.out =  meshSize*meshSize)
# irt_p1inits_d <- matrix(rep(deltaInits, each = Nitems), ncol = meshSize*meshSize)
# irt_p1inits_var <- matrix(rep(1, times = meshSize*meshSize), ncol = meshSize*meshSize)
# # for each starting value, so each trajectory, so each trial, we need nine starting values
# # for the nine items in case 1 but just one starting value for the variance
# irt_p1inits <- matrix(rbind(irt_p1inits_d, irt_p1inits_var), ncol = meshSize*meshSize)
#
# # Obtain trajectories, starting from different initial values.
# irt_trajectories1 <- getTrajectories(irt_p1inits, xIRT, IRT_F1, maxIter = 1000)
#
# # Plot settings.
# par(mar=c(4,4,2,1), cex = 3, font.lab = 2, mgp = c(1.8, 0.5, 0))
#
# item = 9
#
# # Plot of trajectories to fixed point from initial starting positions.
# plot(irt_trajectories1[[1]][item,], rep(0, length(irt_trajectories1[[1]][item,])), pch = 3,
#      xlim = c(-3,3), ylim = c(-0.5,0.5), yaxt='n', type='n',
#      xlab = TeX(paste0('$\\delta_{',item,'}$')), ylab = '', main = NULL)
# for (i in seq(from = 1, to=length(irt_trajectories1), by = 1)){
#     points(irt_trajectories1[[i]][item,1], 0, pch = 18, col = rgb(0,0,0,0.5))
#     arrows(x0 = irt_trajectories1[[i]][item,1:length(irt_trajectories1[[i]][1,])-1], y0 = rep(0, length(irt_trajectories1[[i]][item,])-1),
#            x1 = irt_trajectories1[[i]][item,2:length(irt_trajectories1[[i]][1,])], y1 = rep(0,length(irt_trajectories1[[i]][item,])-1),
#            length = 0.3, angle = 30, code = 2, col=rgb(0,0,0,0.5))
# }
# points(delta[item], 0, col=  rgb(1,0,0,0.7), pch = 18, cex =2)
# text(x =-4.2, y = 0.5, labels = "A", cex = 4, srt=0, xpd = T)
#
# # Plot of the cumulative relative distance traveled.
# plot(cumsum(r(matrix(irt_trajectories1[[1]][item,],nrow=1))), xlim = c(0,60), ylim = c(0,1), type="n",
#      xlab = 't', ylab = TeX('$\\sum_{i=1}^t r(i)$'))
# for (i in seq_along(irt_trajectories1)){
#     lines(cumsum(r(matrix(irt_trajectories1[[i]][item,], nrow=1))),
#           type='o', pch = 16, col = rgb(0,0,0,0.2))
# }
# text(x =-13, y = 1, labels = "B'", cex = 4, srt=0, xpd = T)
#
# # Plot of angles.
# plot(psi(matrix(irt_trajectories1[[1]][item,], nrow=1)), xlim = c(0,60), ylim = c(0, pi), type="n",
#      xlab = 't', yaxt='n', ylab = TeX('$\\psi(t)$'))
# for (i in seq_along(irt_trajectories1)){
#     lines(psi(matrix(irt_trajectories1[[i]][item,], nrow=1)),
#           type='o', pch = 16, col = rgb(0,0,0,0.2))
# }
# axis(side = 2, at=c(0, pi/2, pi), labels = c(0, TeX('$\\pi/2$'),TeX('$\\pi$')))
# text(x =-12, y = pi, labels = "C'", cex = 4, srt=0, xpd = T)
#



# #################
# # IRT Case 2: F #
# #################
#
# # Case 2
# # Initial delta and alpha estimates (for different trajectories).
# deltaInits2 <- seq(from = -3, to = 3, length.out =  meshSize)
# alphaInits2 <- seq(from = 0.01, to = 2, length.out =  meshSize)
# irt_p2inits_d <- matrix(rep(rep(deltaInits2, each = Nitems), meshSize), ncol = meshSize*meshSize)
# irt_p2inits_a <- matrix(rep(alphaInits2, each = Nitems* meshSize), ncol = meshSize*meshSize)
# irt_p2inits_var <- matrix(rep(1, each = meshSize*meshSize), ncol = meshSize*meshSize)
# irt_p2inits <- matrix(rbind(irt_p2inits_d, irt_p2inits_a, irt_p2inits_var),
#                       ncol = meshSize*meshSize)
#
# # Obtain trajectories, starting from different initial values.
# irt_trajectories2 <- getTrajectories(irt_p2inits, xIRT, IRT_F2, maxIter = 1000)
#
# par(mar=c(4,4,2,1), cex = 3, font.lab = 2, mgp = c(1.8, 0.5, 0))
#
# # Plot of trajectories to fixed point from initial starting positions.
# plot(irt_trajectories2[[1]][item,], irt_trajectories2[[1]][Nitems+item,], pch = 3,
#      xlim = c(-3,3), ylim = c(0,2), type='n',
#      xlab = TeX(paste0('$\\delta_{',item,'}$')),
#      ylab = TeX(paste0('$\\alpha_{',item,'}$')), main = NULL)
# for (i in seq(from = 1, to=length(irt_trajectories2), by = 1)){
#     points(irt_trajectories2[[i]][item,1], irt_trajectories2[[i]][Nitems+item,1], pch = 18, col = rgb(0,0,0,0.5))
#     arrows(x0 = irt_trajectories2[[i]][item,1:length(irt_trajectories2[[i]][1,])-1], y0 = irt_trajectories2[[i]][Nitems+item,1:length(irt_trajectories2[[i]][1,])-1],
#            x1 = irt_trajectories2[[i]][item,2:length(irt_trajectories2[[i]][1,])], y1 =  irt_trajectories2[[i]][Nitems+item,2:length(irt_trajectories2[[i]][1,])],
#            length = 0.3, angle = 30, code = 2, rgb(0,0,0,0.5))
# }
# points(delta[item], alpha[item], col=  rgb(1,0,0,0.7), pch = 18, cex =2)
# text(x =-4.3, y = 2, labels = "A'", cex = 4, srt=0, xpd = T)
#
# # Plot of the cumulative relative distance traveled.
# plot(cumsum(r(rbind(irt_trajectories2[[1]][item,], irt_trajectories2[[1]][Nitems+item,]))),
#      xlim = c(0,60), ylim = c(0,1), type="n",
#      xlab = 't', ylab = TeX('$\\sum_{i=1}^t r(i)$'))
# for (i in seq_along(irt_trajectories2)){
#     lines(cumsum(r(rbind(irt_trajectories2[[i]][item,], irt_trajectories2[[i]][Nitems+item,]))),
#           type='o', pch = 16, col = rgb(0,0,0,0.2))
# }
# text(x =-13, y = 1, labels = "B'", cex = 4, srt=0, xpd = T)
#
# # Plot of angles.
# plot(psi(rbind(irt_trajectories2[[1]][item,], irt_trajectories2[[1]][Nitems+item,])),
#      xlim = c(0,60), ylim = c(0, pi), type="n",
#      xlab = 't', yaxt='n', ylab = TeX('$\\psi(t)$'))
# for (i in seq_along(irt_trajectories2)){
#     lines(psi(rbind(irt_trajectories2[[i]][item,], irt_trajectories2[[i]][Nitems+item,])),
#           type='o', pch = 16, col = rgb(0,0,0,0.2))
# }
# axis(side = 2, at=c(0, pi/2, pi), labels = c(0, TeX('$\\pi/2$'),TeX('$\\pi$')))
# text(x =-12, y = pi, labels = "C'", cex = 4, srt=0, xpd = T)

# ################################
# # IRT Case 3: high-dimensional #
# ################################
#
# # Case 3
# # Initial delta and alpha estimates (for different trajectories).
# irt_p3inits_d <- matrix(rep(rep(deltaInits2, each = NitemsMulti), meshSize), ncol = meshSize*meshSize)
# irt_p3inits_a <- matrix(rep(alphaInits2, each = NitemsMulti* meshSize), ncol = meshSize*meshSize)
# irt_p3inits_var <- matrix(rep(1, each = meshSize*meshSize), ncol = meshSize*meshSize)
# irt_p3inits <- matrix(rbind(irt_p3inits_d, irt_p3inits_a, irt_p3inits_var), ncol = meshSize*meshSize)
#
# # Obtain trajectories, starting from different initial values.
# irt_trajectories3 <- getTrajectories(irt_p3inits, xIRTMulti, IRT_F2, maxIter = 1000)
#
# par(mar=c(4,4,2,1), cex = 3, font.lab = 2, mgp = c(1.8, 0.5, 0))
#
# # Plot of trajectories to fixed point from initial starting positions.
# plot(irt_trajectories3[[1]][item,], irt_trajectories3[[1]][NitemsMulti+item,], pch = 3,
#      xlim = c(-3,3), ylim = c(0,2), type='n',
#      xlab = TeX(paste0('$\\delta_{',item,'}$')),
#      ylab = TeX(paste0('$\\alpha_{',item,'}$')), main = NULL)
# for (i in seq(from = 1, to=length(irt_trajectories3), by = 1)){
#     points(irt_trajectories3[[i]][item,1], irt_trajectories3[[i]][NitemsMulti+item,1], pch = 18, col = rgb(0,0,0,0.5))
#     arrows(x0 = irt_trajectories3[[i]][item,1:length(irt_trajectories3[[i]][1,])-1], y0 = irt_trajectories3[[i]][NitemsMulti+item,1:length(irt_trajectories3[[i]][1,])-1],
#            x1 = irt_trajectories3[[i]][item,2:length(irt_trajectories3[[i]][1,])], y1 =  irt_trajectories3[[i]][NitemsMulti+item,2:length(irt_trajectories3[[i]][1,])],
#            length = 0.3, angle = 30, code = 2, rgb(0,0,0,0.5))
# }
# points(delta[item], alpha[item], col = 'red', pch = 18)
# text(x =-4.3, y = 2, labels = "A'", cex = 4, srt=0, xpd = T)
#
# # Plot of the cumulative relative distance traveled.
# plot(cumsum(r(rbind(irt_trajectories3[[1]][item,], irt_trajectories3[[1]][NitemsMulti+item,]))),
#      xlim = c(0,200), ylim = c(0,1), type="n",
#      xlab = 't', ylab = TeX('$\\sum_{i=1}^t r(i)$'))
# for (i in seq_along(irt_trajectories3)){
#     points(cumsum(r(rbind(irt_trajectories3[[i]][item,],
#                           irt_trajectories3[[i]][NitemsMulti+item,]))),
#            type='o', pch = 16, col = rgb(0,0,0,0.2))
# }
# text(x =-45, y = 1, labels = "B'", cex = 4, srt=0, xpd = T)
#
# # Plot of angles.
# plot(psi(rbind(irt_trajectories3[[1]][item,], irt_trajectories3[[1]][NitemsMulti+item,])),
#      xlim = c(0,200), ylim = c(0, pi), type="n",
#      xlab = 't', yaxt='n', ylab = TeX('$\\psi(t)$'))
# for (i in seq_along(irt_trajectories3)){
#     points(psi(rbind(irt_trajectories3[[i]][item,],
#                      irt_trajectories3[[i]][NitemsMulti+item,])),
#            type='o', pch = 16, col = rgb(0,0,0,0.2))
# }
# axis(side = 2, at=c(0, pi/2, pi), labels = c('0', TeX('$\\pi/2$'),TeX('$\\pi$')))
# text(x =-45, y = pi, labels = "C'", cex = 4, srt=0, xpd = T)
#




